@@ -1,39 +1,23 @@
+import itertools
+import joblib
+import warnings
+
 from collections import Counter
 from pathlib import Path
-from typing import Dict, Text, Any, Optional, List, Tuple, NamedTuple, Union, Callable
-
-import joblib
-import numpy as np
-import sklearn_crfsuite
+from typing import Dict, Text, Any, Optional, List, Tuple, Union, Callable
 
 from sklearn_crfsuite import CRF, metrics
 from spacy.language import Language
 from spacy.tokens.doc import Doc
 
-from spacy_crfsuite.bilou import (
-    entity_name_from_tag,
-    bilou_tags_from_offsets,
-    bilou_prefix_from_tag,
-    get_entity_offsets,
-)
-
-from spacy_crfsuite.compat import msg
-from spacy_crfsuite.constants import TOKENS, PATTERN, DENSE_FEATURES
-from spacy_crfsuite.dense_features import DenseFeatures
-from spacy_crfsuite.tokenizer import Token, SpacyTokenizer, Tokenizer
+from spacy_crfsuite.bilou import entity_name_from_tag, bilou_prefix_from_tag, NO_ENTITY_TAG
+from spacy_crfsuite.features import CRFToken, Featurizer
+from spacy_crfsuite.tokenizer import Token, SpacyTokenizer
 from spacy_crfsuite.utils import override_defaults
 
 
-class CRFToken(NamedTuple):
-    text: Text
-    tag: Text
-    entity: Text
-    pattern: Dict[Text, Any]
-    dense_features: np.ndarray
-
-
 class CRFExtractor:
-    defaults = {
+    defaults: Dict[str, Any] = {
         # BILOU_flag determines whether to use BILOU tagging or not.
         # More rigorous however requires more examples per entity
         # rule of thumb: use only if more than 100 egs. per entity
@@ -42,8 +26,6 @@ class CRFExtractor:
         # after holding keys about which features to use for each token,
         # for example, 'title' in array before will have the feature
         # "is the preceding token in title case?"
-        # POS features require SpacyTokenizer
-        # pattern feature require RegexFeaturizer
         "features": [
             ["low", "title", "upper"],
             [
@@ -58,6 +40,7 @@ class CRFExtractor:
                 "title",
                 "digit",
                 "pattern",
+                "shape",
             ],
             ["low", "title", "upper"],
         ],
@@ -87,10 +70,9 @@ class CRFExtractor:
         "pos2": lambda crf_token: crf_token.tag[:2] if crf_token.tag is not None else None,
         "upper": lambda crf_token: crf_token.text.isupper(),
         "digit": lambda crf_token: crf_token.text.isdigit(),
-        "pattern": lambda crf_token: crf_token.pattern
-        if crf_token.pattern is not None
-        else None,
-        "text_dense_features": lambda crf_token: crf_token.dense_features,
+        "shape": lambda crf_token: crf_token.shape,
+        "pattern": lambda crf_token: crf_token.pattern,
+        "dense_features": lambda crf_token: crf_token.dense_features
     }
 
     def __init__(
@@ -101,6 +83,7 @@ class CRFExtractor:
 
         self.component_config = override_defaults(self.defaults, component_config)
         self.ent_tagger = ent_tagger
+        self.featurizer = Featurizer(use_dense_features=self.use_dense_features())
 
     def from_disk(self, path: Union[Path, str] = "model.pkl") -> "CRFExtractor":
         """Load component from disk.
@@ -140,7 +123,7 @@ class CRFExtractor:
             bool
         """
         for feature_list in self.component_config["features"]:
-            if DENSE_FEATURES in feature_list:
+            if "dense_features" in feature_list:
                 return True
         return False
 
@@ -158,22 +141,27 @@ class CRFExtractor:
         """
         self._check_runtime()
 
-        text_data = self._from_text_to_crf(example)
-        features = self._crf_tokens_to_features(text_data)
+        crf_tokens: List[CRFToken] = self.featurizer(example)
+        features: List[Dict[Text, Any]] = self._crf_tokens_to_features(crf_tokens)
         entities = self.ent_tagger.predict_marginals_single(features)
         return self._from_crf_to_json(example, entities)
 
-    def train(self, training_samples: List[List[CRFToken]]) -> "CRFExtractor":
+    def train(
+        self,
+        training_samples: List[List[CRFToken]],
+        dev_samples: Optional[List[List[CRFToken]]] = None,
+    ) -> "CRFExtractor":
         """Train the entity tagger with examples.
 
         Args:
             training_samples (list): list of training examples.
+            dev_samples (list): optional, list of dev examples.
 
         Returns:
             CRFExtractor
         """
         if self.ent_tagger is None:
-            self.ent_tagger = sklearn_crfsuite.CRF(
+            self.ent_tagger = CRF(
                 algorithm=self.component_config["algorithm"],
                 c1=self.component_config["c1"],
                 c2=self.component_config["c2"],
@@ -183,7 +171,15 @@ class CRFExtractor:
 
         X_train = [self._crf_tokens_to_features(sent) for sent in training_samples]
         y_train = [self._crf_tokens_to_tags(sent) for sent in training_samples]
-        self.ent_tagger.fit(X_train, y_train)
+
+        if dev_samples:
+            X_dev = [self._crf_tokens_to_features(sent) for sent in dev_samples]
+            y_dev = [self._crf_tokens_to_tags(sent) for sent in dev_samples]
+        else:
+            X_dev = None
+            y_dev = None
+
+        self.ent_tagger.fit(X_train, y_train, X_dev=X_dev, y_dev=y_dev)
         return self
 
     def eval(self, eval_samples: List[List[CRFToken]]) -> Optional[Tuple[Any, Text]]:
@@ -201,7 +197,7 @@ class CRFExtractor:
         y_test = [self._crf_tokens_to_tags(sent) for sent in eval_samples]
 
         labels = list(self.ent_tagger.classes_)
-        labels.remove("O")
+        labels.remove(NO_ENTITY_TAG)
         sorted_labels = sorted(labels, key=lambda name: (name[1:], name[0]))
 
         y_pred = self.ent_tagger.predict(X_test)
@@ -211,6 +207,58 @@ class CRFExtractor:
         )
         return f1_score, classification_report
 
+    def fine_tune(
+        self,
+        val_samples: List[List[CRFToken]],
+        cv=3,
+        n_iter=10,
+        n_jobs=-1,
+        random_state=None,
+    ):
+        """Fine-tune hyper params (i.e L1/L2 regularization term).
+
+        Args:
+            val_samples (list): samples for cross-validation.
+            cv (int): optional, cross-validation K-fold.
+            n_iter (int): optional, num of iterations.
+            n_jobs (int): optional, num of threads (-1 = ``multiprocessing.cpu_count()``).
+            random_state (int): optional, random seed.
+
+        Returns:
+            a fitted instance of ``RandomizedSearchCV``.
+        """
+        import scipy
+
+        from sklearn.metrics import make_scorer
+        from sklearn.model_selection import RandomizedSearchCV
+
+        params_space = {
+            "c1": scipy.stats.expon(scale=0.5),
+            "c2": scipy.stats.expon(scale=0.05),
+        }
+        crf = CRF(
+            algorithm=self.component_config["algorithm"],
+            max_iterations=self.component_config["max_iter"],
+            all_possible_transitions=self.component_config["all_possible_transitions"],
+        )
+
+        X_train = [self._crf_tokens_to_features(sent) for sent in val_samples]
+        y_train = [self._crf_tokens_to_tags(sent) for sent in val_samples]
+        labels = list(set(itertools.chain.from_iterable(y_train)) - {NO_ENTITY_TAG})
+        f1_scorer = make_scorer(metrics.flat_f1_score, average="weighted", labels=labels)
+        rs = RandomizedSearchCV(
+            crf,
+            params_space,
+            cv=cv,
+            verbose=1,
+            n_jobs=n_jobs,
+            n_iter=n_iter,
+            scoring=f1_scorer,
+            random_state=random_state,
+        )
+        rs.fit(X_train, y_train)
+        return rs
+
     def _check_runtime(self):
         """Helper to check runtime before using component for predictions."""
         if self.ent_tagger is None:
@@ -323,7 +371,7 @@ class CRFExtractor:
     @staticmethod
     def _tokens_without_cls(message: Dict) -> List[Token]:
         # [:-1] to remove the CLS token from the list of tokens
-        return message.get(TOKENS)[:-1]
+        return message.get("tokens")[:-1]
 
     def _find_bilou_end(self, word_idx, entities) -> Any:
         ent_word_idx = word_idx + 1
@@ -335,12 +383,11 @@ class CRFExtractor:
 
         while not finished:
             label, label_confidence = self.most_likely_entity(ent_word_idx, entities)
-
             confidence = min(confidence, label_confidence)
 
             if label[2:] != entity_label:
                 # words are not tagged the same entity class
-                msg and msg.info(
+                warnings.warn(
                     "Inconsistent BILOU tagging found, B- tag, L- "
                     "tag pair encloses multiple entity classes.i.e. "
                     "[B-a, I-b, L-a] instead of [B-a, I-a, L-a].\n"
@@ -357,7 +404,7 @@ class CRFExtractor:
                 # entity not closed by an L- tag
                 finished = True
                 ent_word_idx -= 1
-                msg and msg.info(
+                warnings.warn(
                     "Inconsistent BILOU tagging found, B- tag not "
                     "closed by L- tag, i.e [B-a, I-a, O] instead of "
                     "[B-a, L-a, O].\nAssuming last tag is L-"
@@ -422,7 +469,7 @@ class CRFExtractor:
         for word_idx in range(len(tokens)):
             entity_label, confidence = self.most_likely_entity(word_idx, entities)
             word = tokens[word_idx]
-            if entity_label != "O":
+            if entity_label != NO_ENTITY_TAG:
                 ent = {
                     "start": word.start,
                     "end": word.end,
@@ -484,142 +531,6 @@ class CRFExtractor:
     def _crf_tokens_to_tags(sentence: List[CRFToken]) -> List[Text]:
         return [crf_token.entity for crf_token in sentence]
 
-    def from_json_to_crf(
-        self, message: Dict, entity_offsets: List[Tuple[int, int, Text]]
-    ) -> List[CRFToken]:
-        """Convert json examples to format of underlying crfsuite."""
-
-        tokens = self._tokens_without_cls(message)
-        ents = bilou_tags_from_offsets(tokens, entity_offsets)
-
-        # collect badly annotated examples
-        collected = []
-        for t, e in zip(tokens, ents):
-            if e == "-":
-                collected.append(t)
-            elif collected:
-                collected_text = " ".join([t.text for t in collected])
-                msg and msg.warn(
-                    f"Misaligned entity annotation for '{collected_text}' "
-                    f"in sentence: \"{message['text']}\". "
-                    f"Make sure the start and end values of the "
-                    f"annotated training examples end at token "
-                    f"boundaries (e.g. don't include trailing "
-                    f"whitespaces or punctuation)."
-                )
-                collected = []
-
-        if not self.component_config["BILOU_flag"]:
-            for i, label in enumerate(ents):
-                if bilou_prefix_from_tag(label):  # removes BILOU prefix from label
-                    ents[i] = entity_name_from_tag(label)
-
-        return self._from_text_to_crf(message, ents)
-
-    @staticmethod
-    def __pattern_of_token(message: Dict, i: int) -> Dict:
-        if message.get(TOKENS) is not None:
-            return message.get(TOKENS)[i].get(PATTERN, {})
-        else:
-            return {}
-
-    @staticmethod
-    def __get_dense_features(message: Dict) -> Optional[List[Any]]:
-        features = message.get(DENSE_FEATURES)
-        if features is None:
-            return None
-
-        tokens = message.get(TOKENS, [])
-        if len(tokens) != len(features):
-            msg and msg.warning(
-                f"Number of features ({len(features)}) for attribute "
-                f"'{DENSE_FEATURES}' "
-                f"does not match number of tokens ({len(tokens)}). Set "
-                f"'return_sequence' to true in the corresponding featurizer in order "
-                f"to make use of the features in 'CRFEntityExtractor'."
-            )
-            return None
-
-        # convert to python-crfsuite feature format
-        features_out = []
-        for feature in features:
-            feature_dict = {
-                str(index): token_features for index, token_features in enumerate(feature)
-            }
-            converted = {DENSE_FEATURES: feature_dict}
-            features_out.append(converted)
-        return features_out
-
-    def _from_text_to_crf(
-        self, message: Dict, entities: List[Text] = None
-    ) -> List[CRFToken]:
-        """Takes a sentence and switches it to crfsuite format."""
-        crf_format = []
-        tokens = self._tokens_without_cls(message)
-        text_dense_features = self.__get_dense_features(message)
-        for i, token in enumerate(tokens):
-            pattern = self.__pattern_of_token(message, i)
-            entity = entities[i] if entities else "N/A"
-            tag = token.get("pos")
-            dense_features = (
-                text_dense_features[i] if text_dense_features is not None else []
-            )
-            crf_token = CRFToken(token.text, tag, entity, pattern, dense_features)
-            crf_format.append(crf_token)
-        return crf_format
-
-
-def prepare_example(
-    example: Dict,
-    crf_extractor: CRFExtractor,
-    tokenizer: Optional[Tokenizer] = None,
-    dense_features: Optional[DenseFeatures] = None,
-) -> Optional[List[CRFToken]]:
-    """Translate training example to CRF feature space.
-
-    Args:
-        example (dict): example dict. must have either "doc", "tokens" or "text" field.
-        crf_extractor (CRFExtractor): CRF component.
-        tokenizer (Tokenizer): tokenizer.
-        dense_features (DenseFeatures): dense features.
-
-    Returns:
-        List[CRFToken], CRF example.
-    """
-    if not example:
-        return
-
-    tokenizer = tokenizer or SpacyTokenizer()
-
-    if "tokens" in example:
-        pass
-
-    elif "doc" in example:
-        tokens = tokenizer.tokenize(example, attribute="doc")
-        if isinstance(tokenizer, SpacyTokenizer):
-            tokenizer.add_cls_token(tokens)
-        example["tokens"] = tokens
-
-    elif "text" in example:
-        tokens = tokenizer.tokenize(example, attribute="text")
-        if isinstance(tokenizer, SpacyTokenizer):
-            tokenizer.add_cls_token(tokens)
-        example["tokens"] = tokens
-
-    else:
-        raise ValueError(f"empty example: {example}")
-
-    if dense_features:
-        text_dense_features = dense_features(
-            example, attribute="doc" if "doc" in example else "text"
-        )
-        if len(text_dense_features) > 0:
-            example["text_dense_features"] = text_dense_features
-
-    entity_offsets = get_entity_offsets(example)
-    crf_example = crf_extractor.from_json_to_crf(example, entity_offsets)
-    return crf_example
-
 
 class CRFEntityExtractor(object):
     """spaCy v2.0 pipeline component that sets entity annotations
@@ -629,18 +540,13 @@ class CRFEntityExtractor(object):
     See ```CRFExtractor``` for CRF implementation details.
     """
 
-    name = "crf_entity_extractor"
+    name = "crf_ner"
 
     def __init__(self, nlp: Language, crf_extractor: Optional[CRFExtractor] = None):
         self.nlp = nlp
         self.crf_extractor = crf_extractor
         self.spacy_tokenizer = SpacyTokenizer(nlp)
 
-        if self.crf_extractor.use_dense_features():
-            self.dense_features = DenseFeatures(nlp)
-        else:
-            self.dense_features = None
-
     def __call__(self, doc: Doc):
         """Apply the pipeline component on a Doc object and modify it if matches
         are found. Return the Doc, so it can be processed by the next component
@@ -661,19 +567,14 @@ class CRFEntityExtractor(object):
                 "Did you call `.from_disk()` method ?"
             )
 
-        message = {"doc": doc, "text": doc.text}
-        prepare_example(
-            message,
-            crf_extractor=self.crf_extractor,
-            tokenizer=self.spacy_tokenizer,
-            dense_features=self.dense_features,
-        )
+        example = {"doc": doc, "text": doc.text}
+        self.spacy_tokenizer.tokenize(example, attribute="doc")
 
         spans = [
             doc.char_span(
                 entity_dict["start"], entity_dict["end"], label=entity_dict["entity"]
             )
-            for entity_dict in self.crf_extractor.process(message)
+            for entity_dict in self.crf_extractor.process(example)
         ]
 
         doc.ents = list(doc.ents) + spans
